# Copyright (c) Facebook, Inc. and its affiliates.

import detectron2.utils.comm as comm
import wandb
import os
import numpy as np
from typing import Tuple
from PIL import Image
import torch
from torch import nn
from torch.nn import functional as F

from detectron2.config import configurable
from detectron2.data import MetadataCatalog
from detectron2.modeling import META_ARCH_REGISTRY, build_backbone, build_sem_seg_head
from detectron2.modeling.backbone import Backbone
from detectron2.modeling.postprocessing import sem_seg_postprocess
from detectron2.structures import Boxes, ImageList, Instances, BitMasks
from detectron2.utils.memory import retry_if_cuda_oom

from .modeling.criterion import SetCriterion, ReconCriterion
from .modeling.matcher import HungarianMatcher


@META_ARCH_REGISTRY.register()
class MaskFormerTTT(nn.Module):
    """
    Main class for mask classification semantic segmentation architectures.
    """

    @configurable
    def __init__(
        self,
        *,
        backbone: Backbone,
        sem_seg_head: nn.Module,
        recon_head: nn.Module,
        recon_loss: nn.Module,
        use_recon_head: bool,
        use_seg_head: bool,
        criterion: nn.Module,
        num_queries: int,
        object_mask_threshold: float,
        overlap_threshold: float,
        metadata,
        size_divisibility: int,
        sem_seg_postprocess_before_inference: bool,
        pixel_mean: Tuple[float],
        pixel_std: Tuple[float],
        # inference
        semantic_on: bool,
        panoptic_on: bool,
        instance_on: bool,
        test_topk_per_image: int,
        running_vid: str
    ):
        """
        Args:
            backbone: a backbone module, must follow detectron2's backbone interface
            sem_seg_head: a module that predicts semantic segmentation from backbone features
            criterion: a module that defines the loss
            num_queries: int, number of queries
            object_mask_threshold: float, threshold to filter query based on classification score
                for panoptic segmentation inference
            overlap_threshold: overlap threshold used in general inference for panoptic segmentation
            metadata: dataset meta, get `thing` and `stuff` category names for panoptic
                segmentation inference
            size_divisibility: Some backbones require the input height and width to be divisible by a
                specific integer. We can use this to override such requirement.
            sem_seg_postprocess_before_inference: whether to resize the prediction back
                to original input size before semantic segmentation inference or after.
                For high-resolution dataset like Mapillary, resizing predictions before
                inference will cause OOM error.
            pixel_mean, pixel_std: list or tuple with #channels element, representing
                the per-channel mean and std to be used to normalize the input image
            semantic_on: bool, whether to output semantic segmentation prediction
            instance_on: bool, whether to output instance segmentation prediction
            panoptic_on: bool, whether to output panoptic segmentation prediction
            test_topk_per_image: int, instance segmentation parameter, keep topk instances per image
        """
        super().__init__()
        self.backbone = backbone
        self.sem_seg_head = sem_seg_head
        self.recon_head = recon_head
        self.recon_loss = recon_loss
        self.use_recon_head = use_recon_head
        self.use_seg_head = use_seg_head
        self.criterion = criterion
        self.num_queries = num_queries
        self.overlap_threshold = overlap_threshold
        self.object_mask_threshold = object_mask_threshold
        self.metadata = metadata
        if size_divisibility < 0:
            # use backbone size_divisibility if not set
            size_divisibility = self.backbone.size_divisibility
        self.size_divisibility = size_divisibility
        self.sem_seg_postprocess_before_inference = sem_seg_postprocess_before_inference
        self.register_buffer("pixel_mean", torch.Tensor(pixel_mean).view(-1, 1, 1), False)
        self.register_buffer("pixel_std", torch.Tensor(pixel_std).view(-1, 1, 1), False)

        # additional args
        self.semantic_on = semantic_on
        self.instance_on = instance_on
        self.panoptic_on = panoptic_on
        self.test_topk_per_image = test_topk_per_image

        self.running_vid = running_vid
        # self.wandb_log_videos = ["0000", "0005", "0006"]
        self.wandb_log_videos = ["0000", "0005"]
    
        if self.use_seg_head and not self.use_recon_head:
            # We are probing here, so lets freeze the backbone. 
            print('Freezing the backbone, training only the segmentation head.')
            self.backbone.eval()
            for param in self.backbone.parameters():
                param.requires_grad = False
        elif self.use_recon_head and not self.use_seg_head:
            print('Training only the reconstruction head.')

        # DEBUG
        self.patch_size = self.recon_loss.patch_size

        if not self.semantic_on:
            assert self.sem_seg_postprocess_before_inference

    @classmethod
    def from_config(cls, cfg):
        backbone = build_backbone(cfg)
        sem_seg_head = build_sem_seg_head(cfg, backbone.output_shape())
        
        # import ipdb; ipdb.set_trace()

        running_vid = cfg.OUTPUT_DIR.split('/')[-1].split('_')[0]

        use_recon_head = cfg.TTT.USE_RECON_HEAD
        use_seg_head = cfg.TTT.USE_SEG_HEAD
        cfg_recon = cfg.clone()
        cfg_recon.defrost()
        cfg_recon.MODEL.SEM_SEG_HEAD.NAME = "MaskFormerReconHead"
        cfg_recon.MODEL.SEM_SEG_HEAD.NUM_CLASSES = 3 * (cfg_recon.MODEL.SWIN.PATCH_SIZE ** 2)
        cfg_recon.MODEL.MASK_FORMER.NUM_OBJECT_QUERIES = 48
        cfg_recon.freeze()
        recon_head = build_sem_seg_head(cfg_recon, backbone.output_shape())

        # Loss parameters:
        deep_supervision = cfg.MODEL.MASK_FORMER.DEEP_SUPERVISION
        no_object_weight = cfg.MODEL.MASK_FORMER.NO_OBJECT_WEIGHT

        # loss weights
        class_weight = cfg.MODEL.MASK_FORMER.CLASS_WEIGHT
        dice_weight = cfg.MODEL.MASK_FORMER.DICE_WEIGHT
        mask_weight = cfg.MODEL.MASK_FORMER.MASK_WEIGHT

        # building criterion
        matcher = HungarianMatcher(
            cost_class=class_weight,
            cost_mask=mask_weight,
            cost_dice=dice_weight,
            num_points=cfg.MODEL.MASK_FORMER.TRAIN_NUM_POINTS,
        )

        weight_dict = {"loss_ce": class_weight, "loss_mask": mask_weight, "loss_dice": dice_weight}

        if deep_supervision:
            dec_layers = cfg.MODEL.MASK_FORMER.DEC_LAYERS
            aux_weight_dict = {}
            for i in range(dec_layers - 1):
                aux_weight_dict.update({k + f"_{i}": v for k, v in weight_dict.items()})
            weight_dict.update(aux_weight_dict)

        losses = ["labels", "masks"]

        criterion = SetCriterion(
            sem_seg_head.num_classes,
            matcher=matcher,
            weight_dict=weight_dict,
            eos_coef=no_object_weight,
            losses=losses,
            num_points=cfg.MODEL.MASK_FORMER.TRAIN_NUM_POINTS,
            oversample_ratio=cfg.MODEL.MASK_FORMER.OVERSAMPLE_RATIO,
            importance_sample_ratio=cfg.MODEL.MASK_FORMER.IMPORTANCE_SAMPLE_RATIO,
        )

        recon_criterion = ReconCriterion(patch_size=cfg.MODEL.SWIN.PATCH_SIZE, recon_loss_weight=cfg.TTT.RECON_LOSS_WEIGHT)

        return {
            "backbone": backbone,
            "sem_seg_head": sem_seg_head,
            "recon_head": recon_head,
            "recon_loss": recon_criterion,
            "use_recon_head" : use_recon_head,
            "use_seg_head" : use_seg_head,
            "criterion": criterion,
            "num_queries": cfg.MODEL.MASK_FORMER.NUM_OBJECT_QUERIES,
            "object_mask_threshold": cfg.MODEL.MASK_FORMER.TEST.OBJECT_MASK_THRESHOLD,
            "overlap_threshold": cfg.MODEL.MASK_FORMER.TEST.OVERLAP_THRESHOLD,
            "metadata": MetadataCatalog.get(cfg.DATASETS.TRAIN[0]),
            "size_divisibility": cfg.MODEL.MASK_FORMER.SIZE_DIVISIBILITY,
            "sem_seg_postprocess_before_inference": (
                cfg.MODEL.MASK_FORMER.TEST.SEM_SEG_POSTPROCESSING_BEFORE_INFERENCE
                or cfg.MODEL.MASK_FORMER.TEST.PANOPTIC_ON
                or cfg.MODEL.MASK_FORMER.TEST.INSTANCE_ON
            ),
            "pixel_mean": cfg.MODEL.PIXEL_MEAN,
            "pixel_std": cfg.MODEL.PIXEL_STD,
            # inference
            "semantic_on": cfg.MODEL.MASK_FORMER.TEST.SEMANTIC_ON,
            "instance_on": cfg.MODEL.MASK_FORMER.TEST.INSTANCE_ON,
            "panoptic_on": cfg.MODEL.MASK_FORMER.TEST.PANOPTIC_ON,
            "test_topk_per_image": cfg.TEST.DETECTIONS_PER_IMAGE,
            "running_vid" : running_vid
        }

    @property
    def device(self):
        return self.pixel_mean.device

    def forward(self, batched_inputs):
        # import ipdb; ipdb.set_trace()

        """
        Args:
            batched_inputs: a list, batched outputs of :class:`DatasetMapper`.
                Each item in the list contains the inputs for one image.
                For now, each item in the list is a dict that contains:
                   * "image": Tensor, image in (C, H, W) format.
                   * "instances": per-region ground truth
                   * Other information that's included in the original dicts, such as:
                     "height", "width" (int): the output resolution of the model (may be different
                     from input resolution), used in inference.
        Returns:
            list[dict]:
                each dict has the results for one image. The dict contains the following keys:

                * "sem_seg":
                    A Tensor that represents the
                    per-pixel segmentation prediced by the head.
                    The prediction has shape KxHxW that represents the logits of
                    each class for each pixel.
                * "panoptic_seg":
                    A tuple that represent panoptic output
                    panoptic_seg (Tensor): of shape (height, width) where the values are ids for each segment.
                    segments_info (list[dict]): Describe each segment in `panoptic_seg`.
                        Each dict contains keys "id", "category_id", "isthing".
        """
        images = [x["image"].to(self.device) for x in batched_inputs]
        images = [(x - self.pixel_mean) / self.pixel_std for x in images]
        images = ImageList.from_tensors(images, self.size_divisibility)
        img = images.tensor


        # import ipdb; ipdb.set_trace()
        if self.training:
            all_losses = {}

            if self.use_seg_head:
                # Concatenate 4th channel of all 1s
                mask = torch.ones((img.shape[0], 1, img.shape[-2], img.shape[-1])).to(device=img.device, dtype=img.dtype)
                input_tensor = torch.cat([img, mask], dim=1)
                
                features = self.backbone(input_tensor)
                outputs = self.sem_seg_head(features)

                # mask classification target
                if "instances" in batched_inputs[0]:
                    gt_instances = [x["instances"].to(self.device) for x in batched_inputs]
                    targets = self.prepare_targets(gt_instances, images)
                else:
                    targets = None
                
                # import ipdb; ipdb.set_trace()

                # bipartite matching-based loss
                losses = self.criterion(outputs, targets)

                for k in list(losses.keys()):
                    if k in self.criterion.weight_dict:
                        losses[k] *= self.criterion.weight_dict[k]
                    else:
                        # remove this loss if not specified in `weight_dict`
                        losses.pop(k)
                
                all_losses.update(losses)
            
            if self.use_recon_head:
                # Apply mask + concatenate --> recon_tensor
                masks = [x["image_mask"].to(self.device) for x in batched_inputs]
                masks = ImageList.from_tensors(masks, self.size_divisibility)
                recon_mask = masks.tensor

                # import ipdb; ipdb.set_trace()

                # # DEBUGGING ON SINGLE IMAGE
                # recon_mask = torch.stack([x["image_mask"].to(self.device) for x in batched_inputs], dim=0)

                # import ipdb; ipdb.set_trace()

                input_img = img.float() * recon_mask.float()
                masked_img = torch.cat([input_img, recon_mask], dim=1)

                # reconstruction loss
                features = self.backbone(masked_img)
                recon_output = self.recon_head(features)

                # Get semantic-level reconstruction outputs
                mask_pred_results = recon_output["pred_masks"]

                # mask_pred_results: (B, 4, H, W)
                # images.tensor: (B, 4, H, W)
                # import ipdb; ipdb.set_trace()
                recon_losses, alt_losses = self.recon_loss(mask_pred_results, img, recon_mask)
                all_losses.update(recon_losses)

                # # Wandb to save
                # if comm.is_main_process():
                #     if self.running_vid in self.wandb_log_videos:
                #         img_res_dict = self.wandb_logger(input_img, masked_img, mask_pred_results)
                #     else:
                #         img_res_dict = {}
                #     img_res_dict.update(recon_losses)
                #     # img_res_dict.update(alt_losses)
                #     wandb.log(img_res_dict)

            return all_losses
        else:
            # Concatenate 4th channel of all 1s
            mask = torch.ones((img.shape[0], 1, img.shape[-2], img.shape[-1])).to(device=img.device, dtype=img.dtype)
            input_tensor = torch.cat([img, mask], dim=1)
            features = self.backbone(input_tensor)
            outputs = self.sem_seg_head(features)

            mask_cls_results = outputs["pred_logits"]
            mask_pred_results = outputs["pred_masks"]
            # upsample masks
            mask_pred_results = F.interpolate(
                mask_pred_results,
                size=(images.tensor.shape[-2], images.tensor.shape[-1]),
                mode="bilinear",
                align_corners=False,
            )

            del outputs

            # import ipdb; ipdb.set_trace()

            processed_results = []
            for mask_cls_result, mask_pred_result, input_per_image, image_size in zip(
                mask_cls_results, mask_pred_results, batched_inputs, images.image_sizes
            ):
                height = input_per_image.get("height", image_size[0])
                width = input_per_image.get("width", image_size[1])
                processed_results.append({})

                if self.sem_seg_postprocess_before_inference:
                    mask_pred_result = retry_if_cuda_oom(sem_seg_postprocess)(
                        mask_pred_result, image_size, height, width
                    )
                    mask_cls_result = mask_cls_result.to(mask_pred_result)
                
                # semantic segmentation inference
                if self.semantic_on:
                    r = retry_if_cuda_oom(self.semantic_inference)(mask_cls_result, mask_pred_result)
                    if not self.sem_seg_postprocess_before_inference:
                        r = retry_if_cuda_oom(sem_seg_postprocess)(r, image_size, height, width)
                    processed_results[-1]["sem_seg"] = r
                
                # panoptic segmentation inference
                if self.panoptic_on:
                    panoptic_r = retry_if_cuda_oom(self.panoptic_inference)(mask_cls_result, mask_pred_result)
                    processed_results[-1]["panoptic_seg"] = panoptic_r
                
                # instance segmentation inference
                if self.instance_on:
                    instance_r = retry_if_cuda_oom(self.instance_inference)(mask_cls_result, mask_pred_result)
                    processed_results[-1]["instances"] = instance_r

            # import ipdb; ipdb.set_trace()

            # Apply mask + concatenate --> recon_tensor
            if "image_mask" in batched_inputs[0]:
                masks = [x["image_mask"].to(self.device) for x in batched_inputs]
                masks = ImageList.from_tensors(masks, self.size_divisibility)
                recon_mask = masks.tensor

                input_img = img.float() * recon_mask.float()
                masked_img = torch.cat([input_img, recon_mask], dim=1)

                # reconstruction loss
                features = self.backbone(masked_img)
                recon_output = self.recon_head(features)
                recon_pred_results = recon_output["pred_masks"]
                for i, (recon_pred_result, input_per_image, image_size) in enumerate(zip(
                        recon_pred_results, batched_inputs, images.image_sizes)):
                    height = input_per_image.get("height", image_size[0])
                    width = input_per_image.get("width", image_size[1])

                    # Reshape recon to 3-channel image
                    p = self.patch_size
                    h, w = img.shape[-2:]
                    assert (h // p) == recon_pred_result.shape[-2] and (w // p) == recon_pred_result.shape[-1], 'reconstruction input/output mismatch; check patch size'
                    
                    recon = recon_pred_result.reshape(shape=(3, p, p, h // p, w // p))
                    recon = torch.einsum('cpqhw->chpwq', recon)
                    recon = recon.reshape(shape=(3, h, w))

                    # Undo normalization in reconstruction
                    recon = (recon * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]).clamp(min=0, max=255)

                    r = retry_if_cuda_oom(sem_seg_postprocess)(recon, image_size, height, width)
                    processed_results[i]["recon"] = r

            return processed_results


    # def debug_recon_loss(self, images, masks, reconstruction, save_dir):
    def debug_recon_loss(self, images, reconstruction, save_dir):
        # images: torch.Size([B, 4, 512, 1024])
        # masks: torch.Size([B, 1, 512, 1024])
        # reconstruction: torch.Size([4, 48, 128, 256])
        os.makedirs(save_dir, exist_ok=True)

        # for i, (image, mask, recon) in enumerate(zip(images, masks, reconstruction)):
        for i, (image, recon) in enumerate(zip(images, reconstruction)):
            # First 3 channels of RGB
            image_rgb = image[:-1, :, :]
            # Undo normalization in images
            image_rgb = image_rgb * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]
            save_path = os.path.join(save_dir, str(i) + '_rgb.png')
            Image.fromarray(image_rgb.permute(1, 2, 0).detach().cpu().numpy().astype(np.uint8)).save(save_path)

            # # Assert last channel of image == mask
            # image_mask = image[-1, :, :]
            # assert torch.sum(image_mask - mask) < 1e-10, 'Concatenated mask and raw mask do not agree'
            # save_path = os.path.join(save_dir, str(i) + '_mask.png')
            # Image.fromarray((image_mask * 255).detach().cpu().numpy().astype(np.uint8)).save(save_path)
            
            # Reshape recon to 3-channel image
            p = self.patch_size
            h, w = image.shape[-2:]
            assert (h // p) == recon.shape[-2] and (w // p) == recon.shape[-1], 'reconstruction input/output mismatch; check patch size'
            
            recon = recon.reshape(shape=(3, p, p, h // p, w // p))
            recon = torch.einsum('cpqhw->chpwq', recon)
            recon = recon.reshape(shape=(3, h, w))

            # Undo normalization in reconstruction
            recon = (recon * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]).clamp(min=0, max=255)
            save_path = os.path.join(save_dir, str(i) + '_recon.png')
            Image.fromarray(recon.permute(1, 2, 0).detach().cpu().numpy().astype(np.uint8)).save(save_path)


    def wandb_logger(self, input_images, images, reconstruction):
        choice = np.random.choice(np.arange(len(images)))
        raw_image = input_images[choice]
        image = images[choice]
        recon = reconstruction[choice]

        # Raw image
        # Undo normalization in images
        input_image = raw_image * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]
        input_image = Image.fromarray(input_image.permute(1, 2, 0).detach().cpu().numpy().astype(np.uint8))

        # Masked image
        image_rgb = image[:-1, :, :]
        # Undo normalization in images
        image_rgb = image_rgb * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]
        image_rgb = Image.fromarray(image_rgb.permute(1, 2, 0).detach().cpu().numpy().astype(np.uint8))

        # Assert last channel of image == mask
        image_mask = image[-1, :, :]
        # assert torch.sum(image_mask - mask) < 1e-10, 'Concatenated mask and raw mask do not agree'
        # save_path = os.path.join(save_dir, str(i) + '_mask.png')
        mask = Image.fromarray((image_mask * 255).detach().cpu().numpy().astype(np.uint8))
        
        # Reshape recon to 3-channel image
        p = self.patch_size
        h, w = image.shape[-2:]
        assert (h // p) == recon.shape[-2] and (w // p) == recon.shape[-1], 'reconstruction input/output mismatch; check patch size'
        
        recon = recon.reshape(shape=(3, p, p, h // p, w // p))
        recon = torch.einsum('cpqhw->chpwq', recon)
        recon = recon.reshape(shape=(3, h, w))

        # Undo normalization in reconstruction
        recon = (recon * self.pixel_std[:3, :, :] + self.pixel_mean[:3, :, :]).clamp(min=0, max=255)
        recon = Image.fromarray(recon.permute(1, 2, 0).detach().cpu().numpy().astype(np.uint8))

        return {
                "image" : wandb.Image(input_image),
                "masked image": wandb.Image(image_rgb), 
                "reconstruction": wandb.Image(recon),
                # "segmentation" : wandb.Image(segmentation),
                "mask" : wandb.Image(mask)
        }

    def prepare_targets(self, targets, images):
        h_pad, w_pad = images.tensor.shape[-2:]
        new_targets = []
        for targets_per_image in targets:
            # pad gt
            gt_masks = targets_per_image.gt_masks
            padded_masks = torch.zeros((gt_masks.shape[0], h_pad, w_pad), dtype=gt_masks.dtype, device=gt_masks.device)
            padded_masks[:, : gt_masks.shape[1], : gt_masks.shape[2]] = gt_masks
            new_targets.append(
                {
                    "labels": targets_per_image.gt_classes,
                    "masks": padded_masks,
                }
            )
        return new_targets

    def semantic_inference(self, mask_cls, mask_pred):
        mask_cls = F.softmax(mask_cls, dim=-1)[..., :-1]
        mask_pred = mask_pred.sigmoid()
        semseg = torch.einsum("qc,qhw->chw", mask_cls, mask_pred)
        return semseg
    

    def panoptic_inference(self, mask_cls, mask_pred):
        scores, labels = F.softmax(mask_cls, dim=-1).max(-1)
        mask_pred = mask_pred.sigmoid()

        keep = labels.ne(self.sem_seg_head.num_classes) & (scores > self.object_mask_threshold)
        cur_scores = scores[keep]
        cur_classes = labels[keep]
        cur_masks = mask_pred[keep]
        cur_mask_cls = mask_cls[keep]
        cur_mask_cls = cur_mask_cls[:, :-1]

        cur_prob_masks = cur_scores.view(-1, 1, 1) * cur_masks

        h, w = cur_masks.shape[-2:]
        panoptic_seg = torch.zeros((h, w), dtype=torch.int32, device=cur_masks.device)
        segments_info = []

        current_segment_id = 0

        if cur_masks.shape[0] == 0:
            # We didn't detect any mask :(
            return panoptic_seg, segments_info
        else:
            # take argmax
            cur_mask_ids = cur_prob_masks.argmax(0)
            stuff_memory_list = {}
            for k in range(cur_classes.shape[0]):
                pred_class = cur_classes[k].item()
                isthing = pred_class in self.metadata.thing_dataset_id_to_contiguous_id.values()
                mask_area = (cur_mask_ids == k).sum().item()
                original_area = (cur_masks[k] >= 0.5).sum().item()
                mask = (cur_mask_ids == k) & (cur_masks[k] >= 0.5)

                if mask_area > 0 and original_area > 0 and mask.sum().item() > 0:
                    if mask_area / original_area < self.overlap_threshold:
                        continue

                    # merge stuff regions
                    if not isthing:
                        if int(pred_class) in stuff_memory_list.keys():
                            panoptic_seg[mask] = stuff_memory_list[int(pred_class)]
                            continue
                        else:
                            stuff_memory_list[int(pred_class)] = current_segment_id + 1

                    current_segment_id += 1
                    panoptic_seg[mask] = current_segment_id

                    segments_info.append(
                        {
                            "id": current_segment_id,
                            "isthing": bool(isthing),
                            "category_id": int(pred_class),
                        }
                    )

            return panoptic_seg, segments_info

    def instance_inference(self, mask_cls, mask_pred):
        # mask_pred is already processed to have the same shape as original input
        image_size = mask_pred.shape[-2:]

        # [Q, K]
        scores = F.softmax(mask_cls, dim=-1)[:, :-1]
        labels = torch.arange(self.sem_seg_head.num_classes, device=self.device).unsqueeze(0).repeat(self.num_queries, 1).flatten(0, 1)
        # scores_per_image, topk_indices = scores.flatten(0, 1).topk(self.num_queries, sorted=False)
        scores_per_image, topk_indices = scores.flatten(0, 1).topk(self.test_topk_per_image, sorted=False)
        labels_per_image = labels[topk_indices]

        topk_indices = topk_indices // self.sem_seg_head.num_classes
        # mask_pred = mask_pred.unsqueeze(1).repeat(1, self.sem_seg_head.num_classes, 1).flatten(0, 1)
        mask_pred = mask_pred[topk_indices]

        # if this is panoptic segmentation, we only keep the "thing" classes
        if self.panoptic_on:
            keep = torch.zeros_like(scores_per_image).bool()
            for i, lab in enumerate(labels_per_image):
                keep[i] = lab in self.metadata.thing_dataset_id_to_contiguous_id.values()

            scores_per_image = scores_per_image[keep]
            labels_per_image = labels_per_image[keep]
            mask_pred = mask_pred[keep]

        result = Instances(image_size)
        # mask (before sigmoid)
        result.pred_masks = (mask_pred > 0).float()
        result.pred_boxes = Boxes(torch.zeros(mask_pred.size(0), 4))
        # Uncomment the following to get boxes from masks (this is slow)
        # result.pred_boxes = BitMasks(mask_pred > 0).get_bounding_boxes()

        # calculate average mask prob
        mask_scores_per_image = (mask_pred.sigmoid().flatten(1) * result.pred_masks.flatten(1)).sum(1) / (result.pred_masks.flatten(1).sum(1) + 1e-6)
        result.scores = scores_per_image * mask_scores_per_image
        result.pred_classes = labels_per_image
        return result

